# import glob
# import os
# import re
# import shutil

# import discretisedfield as df
# import micromagneticmodel as mm
# import ubermagtable as ut

# import mumax3c as mc


# def oxs_class(term):
#     """Extract the OOMMF ``Oxs_`` class name of an individual term."""
#     mif = getattr(mc.scripts.energy, f"{term.name}_script")(term)
#     return re.search(r"Oxs_([\w_]+)", mif).group(1)


# def schedule_script(func):
#     """Generate OOMMF ``Schedule...`` line for saving an individual value."""
#     if func.__name__ == "energy":
#         return ""  # Datatable with energies is saved by default.
#     elif func.__name__ == "effective_field":
#         if isinstance(func.__self__, mm.Energy):
#             output = "Oxs_RungeKuttaEvolve:evolver:Total field"
#         else:
#             output = f"Oxs_{oxs_class(func.__self__)}::Field"
#     elif func.__name__ == "density":
#         if isinstance(func.__self__, mm.Energy):
#             output = "Oxs_RungeKuttaEvolve:evolver:Total energy density"
#         else:
#             output = f"Oxs_{oxs_class(func.__self__)}::Energy density"
#     else:
#         msg = f"Computing the value of {func} is not supported."
#         raise ValueError(msg)

#     return 'Schedule "{}" archive Step 1\n'.format(output)


def compute(func, system):
    raise NotImplementedError()
    # """Computes a particular value of an energy term or energy container
    # (``energy``, ``density``, or ``effective_field``).

    # Parameters
    # ----------
    # func : callable

    #     A property of an energy term or an energy container.

    # system : micromagneticmodel.System

    #     Micromagnetic system for which the property is calculated.

    # Returns
    # -------
    # numbers.Real, discretisedfield.Field

    #     Resulting value.

    # Examples
    # --------
    # 1. Computing values of energy terms.

    # >>> import micromagneticmodel as mm
    # >>> import oommfc as oc
    # ...
    # >>> system = mm.examples.macrospin()
    # >>> mc.compute(system.energy.zeeman.energy, system)
    # Running OOMMF...
    # -8.8...e-22
    # >>> mc.compute(system.energy.effective_field, system)
    # Running OOMMF...
    # Field(...)
    # >>> mc.compute(system.energy.density, system)
    # Running OOMMF...
    # Field(...)

    # """
    # td = mc.TimeDriver(total_iteration_limit=1)
    # td.drive(
    #     system, t=1e-25, n=1, save=True, overwrite=True, compute=schedule_script(func)
    # )

    # if func.__name__ == "energy":
    #     extension = "*.odt"
    # elif func.__name__ == "effective_field":
    #     extension = "*.ohf"
    # elif func.__name__ == "density":
    #     extension = "*.oef"

    # dirname = os.path.join(system.name, f"compute-{system.drive_number}")
    # output_file = max(
    #     glob.iglob(os.path.join(dirname, extension)), key=os.path.getctime
    # )

    # if func.__name__ == "energy":
    #     table = ut.read(output_file, rename=False)
    #     if isinstance(func.__self__, mm.Energy):
    #         output = table["RungeKuttaEvolve:evolver:Total energy"][0]
    #     else:
    #         output = table[f"{oxs_class(func.__self__)}::Energy"][0]
    # else:
    #     output = df.Field.fromfile(output_file)

    # # Delete "compute" directory after the data is extracted.
    # shutil.rmtree(dirname)

    # # Delete the parent directory if it remains empty after deleting "compute"
    # # directory.
    # if not os.listdir(system.name):
    #     shutil.rmtree(system.name)

    # return output



import pathlib

import mumax3c as mc


def magnetisation_script(system, ovf_format="bin4", abspath=True):
    system.m.orientation.to_file("m0.omf", representation=ovf_format)
    m0_path = pathlib.Path("m0.omf")
    if abspath:
        m0_path = m0_path.absolute().as_posix()  # '/' as path separator required
    mx3 = "// Magnetisation\n"
    mx3 += f'm.LoadFile("{m0_path}")\n'
    mx3 += mc.scripts.mumax3_regions(system, ovf_format=ovf_format, abspath=abspath)
    return mx3



# import mumax3c as mc


def mesh_script(system):
    mx3 = "// Mesh\n"
    if any(i in system.m.mesh.bc for i in "xyz"):  # are there PBC?
        repetitions = [0, 0, 0]  # should be generalised in the future
        for direction in system.m.mesh.bc:
            # Need to figure out the way of setting up the repetitions.
            repetitions[system.m.mesh.region._dim2index(direction)] = 1
        mx3 += "SetPBC({}, {}, {})\n".format(*repetitions)
    mx3 += "SetGridSize({}, {}, {})\n".format(*system.m.mesh.n)
    mx3 += "SetCellSize({}, {}, {})\n\n".format(*system.m.mesh.cell)
    # mx3 += mc.scripts.set_subregions(system)

    return mx3



import os
import shutil


def delete(system):
    """Deletes micromagnetic system files.

    After the system is driven, if ``save=True`` is passed, the files obtained
    from a mumax3 run are saved in the current directory. ``oommfc.delete`` is
    a convenience function for deleting all of them. More precisely, the
    directory with name the same as ``system.name`` is deleted.

    Parameters
    ----------
    system : micromagneticmodel.System

        System whose files are deleted.

    Raises
    ------
    FileNotFoundError

        If the directory with ``system.name`` does not exist.

    Examples
    --------
    1. Delete system files.

    >>> import os
    >>> import mumax3c as calc
    >>> import micromagneticmodel as mm
    ...
    >>> system = mm.examples.macrospin()
    >>> td = calc.TimeDriver()
    >>> td.drive(system, t=1e-12, n=5, save=True)
    Running mumax3...
    >>> os.path.exists(system.name)
    True
    >>> calc.delete(system)  # deletes directory
    >>> os.path.exists(system.name)
    False

    """
    if os.path.exists(system.name):
        shutil.rmtree(system.name)
    else:
        msg = f"Directory {system.name} does not exist."
        raise FileNotFoundError(msg)



from .driver import Driver


class TimeDriver(Driver):
    """Time driver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Mumax3`` documentation (https://mumax.github.io).

    Examples
    --------
    1. Defining driver with a keyword argument.

    >>> import mumax3c as mc
    ...
    >>> td = mc.TimeDriver(DemagAccuracy=6)

    2. Passing an argument which is not allowed.

    >>> import mumax3c as mc
    ...
    >>> td = mc.TimeDriver(myarg=1)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import mumax3c as mc
    ...
    >>> td = mc.TimeDriver()
    >>> td._allowed_attributes
    [...]

    """

    _allowed_attributes = [
        "DemagAccuracy",
        "dt",
        "FixDt",
        "Headroom",
        "LastErr",
        "MaxDt",
        "MaxErr",
        "MinDt",
        "NEval",
        "PeakErr",
        "step",
        "t",
    ]

    def _checkargs(self, **kwargs):
        t, n = kwargs["t"], kwargs["n"]
        if t <= 0:
            msg = f"Cannot drive with {t=}."
            raise ValueError(msg)
        if not isinstance(n, int):
            msg = f"Cannot drive with {type(n)=}."
            raise ValueError(msg)
        if n <= 0:
            msg = f"Cannot drive with {n=}."
            raise ValueError(msg)

    @property
    def _x(self):
        return "t"



import discretisedfield as df
import micromagneticmodel as mm
import numpy as np

import mumax3c as mc


def driver_script(driver, system, compute=None, ovf_format="bin4", **kwargs):
    mx3 = "tableadd(E_total)\n"
    mx3 += "tableadd(dt)\n"
    mx3 += "tableadd(maxtorque)\n"
    if isinstance(driver, mc.MinDriver):
        for attr, value in driver:
            if attr != "evolver":
                mx3 += f"{attr} = {value}\n"

        mx3 += "minimize()\n\n"
        mx3 += "save(m_full)\n"
        mx3 += "tablesave()\n\n"

    if isinstance(driver, mc.RelaxDriver):
        if not system.dynamics.get(type=mm.Damping):
            raise ValueError("A damping term is needed.")
        alpha = system.dynamics.get(type=mm.Damping)[0].alpha
        mx3 += f"alpha = {alpha}\n"

        for attr, value in driver:
            if attr != "evolver":
                mx3 += f"{attr} = {value}\n"

        mx3 += "relax()\n\n"
        mx3 += "save(m_full)\n"
        mx3 += "tablesave()\n\n"

    if isinstance(driver, mc.TimeDriver):
        # Extract dynamics equation parameters.
        gamma0 = (
            precession[0].gamma0
            if (precession := system.dynamics.get(type=mm.Precession))
            else 0
        )
        if system.dynamics.get(type=mm.Damping):
            alpha = system.dynamics.damping.alpha
        else:
            alpha = 0

        mx3 += f"alpha = {alpha}\n"
        if not gamma0:
            mx3 += "doprecess = false\n"
        else:
            mx3 += f"gammaLL = {gamma0/mm.consts.mu0}\n"
            mx3 += "doprecess = true\n"

        if system.dynamics.get(type=mm.ZhangLi):
            (zh_li_term,) = system.dynamics.get(type=mm.ZhangLi)
            u = (
                zh_li_term.u
                if isinstance(zh_li_term.u, df.Field)
                else df.Field(
                    mesh=system.m.mesh,
                    nvdim=3,
                    value=(1.0, 0.0, 0.0),
                    norm=zh_li_term.u,
                )
            )

            mu_B = mm.consts.e * mm.consts.hbar / (2.0 * mm.consts.me)

            j = -np.multiply(
                u * 2 * (1 + zh_li_term.beta**2) * mm.consts.e / (mm.consts.g * mu_B),
                system.m.norm,
            )
            j.to_file("j.ovf", representation=ovf_format)
            mx3 += f"Xi = {zh_li_term.beta}\n"
            mx3 += "Pol = 1\n"  # Current polarization is 1.
            mx3 += 'J.add(LoadFile("j.ovf"), 1)\n'  # 1 means constant in time.

        mx3 += "setsolver(5)\n"
        mx3 += "fixDt = 0\n\n"

        t, n = kwargs["t"], kwargs["n"]

        mx3 += f"for snap_counter:=0; snap_counter<{n}; snap_counter++{{\n"
        mx3 += f"    run({t/n})\n"
        mx3 += "    save(m_full)\n"
        mx3 += "    tablesave()\n"
        mx3 += "}"

    return mx3



import discretisedfield as df
import micromagneticmodel as mm
import numpy as np
import pytest

import mumax3c as mc


def test_mumax3_regions__no_subregion():
    mesh = df.Mesh(p1=(0, 0, 0), p2=(2, 2, 2), cell=(1, 1, 1))
    system = mm.System(name="test")
    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=1)

    subregion_values, subregions_dict = mc.scripts.util._identify_subregions(system)
    assert np.allclose(subregion_values, 0)
    assert len(subregions_dict) == 1

    mc.scripts.mumax3_regions(system)
    subregions = df.Field.from_file("mumax3_regions.omf")
    assert np.allclose(subregions.array, 0.0)
    assert hasattr(system, "region_relator")
    assert system.region_relator == {"": [0]}


def test_mumax3_regions__two_subregions():
    subregions = {
        "r1": df.Region(p1=(0, 0, 0), p2=(2, 2, 1)),
        "r2": df.Region(p1=(0, 0, 1), p2=(2, 2, 2)),
    }
    mesh = df.Mesh(p1=(0, 0, 0), p2=(2, 2, 2), cell=(1, 1, 1), subregions=subregions)
    system = mm.System(name="test")
    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=1)

    subregion_values, subregions_dict = mc.scripts.util._identify_subregions(system)
    assert subregion_values[0, 0, 0, 0] == 1
    assert subregion_values[1, 1, 1, 0] == 2
    assert len(subregions_dict) == 3
    assert subregions_dict == {0: "", 1: "r1", 2: "r2"}

    mc.scripts.util.mumax3_regions(system)
    subregions = df.Field.from_file("mumax3_regions.omf")
    assert np.allclose(np.unique(subregions.array), [0.0, 1.0])
    assert hasattr(system, "region_relator")
    assert system.region_relator == {"": [], "r1": [0], "r2": [1]}


def test_mumax3_regions__two_subregions_gap_ms():
    subregions = {
        "r1": df.Region(p1=(0, 0, 0), p2=(2, 2, 1)),
        "r2": df.Region(p1=(0, 0, 2), p2=(2, 2, 3)),
    }
    mesh = df.Mesh(p1=(0, 0, 0), p2=(2, 2, 3), cell=(1, 1, 1), subregions=subregions)
    system = mm.System(name="test")

    def ms_fun(pos):
        x, _, _ = pos
        return x

    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=ms_fun)

    subregion_values, subregions_dict = mc.scripts.util._identify_subregions(system)
    assert subregion_values[0, 0, 0, 0] == 1
    assert subregion_values[1, 1, 1, 0] == 0
    assert subregion_values[1, 1, 2, 0] == 2
    assert len(subregions_dict) == 3
    assert subregions_dict == {0: "", 1: "r1", 2: "r2"}

    mc.scripts.util.mumax3_regions(system)
    subregions = df.Field.from_file("mumax3_regions.omf")
    assert np.allclose(np.unique(subregions.array), [0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
    assert hasattr(system, "region_relator")
    assert system.region_relator == {"": [0, 1], "r1": [2, 3], "r2": [4, 5]}

    def ms_fun(pos):
        x, _, _ = pos
        if x < 1:
            return 0
        return 1

    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=ms_fun)

    mc.scripts.util.mumax3_regions(system)
    subregions = df.Field.from_file("mumax3_regions.omf")
    assert np.allclose(np.unique(subregions.array), [0, 1, 2, 255])
    assert hasattr(system, "region_relator")
    assert system.region_relator == {"": [0], "r1": [1], "r2": [2]}


def test_identify_subregions__two_overlaping_subregions():
    subregions = {
        "r1": df.Region(p1=(0, 0, 0), p2=(2, 2, 1)),
        "r2": df.Region(p1=(0, 0, 0), p2=(2, 2, 2)),
    }
    mesh = df.Mesh(p1=(0, 0, 0), p2=(2, 2, 2), cell=(1, 1, 1), subregions=subregions)
    system = mm.System(name="test")
    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=1)

    subregion_values, subregions_dict = mc.scripts.util._identify_subregions(system)
    assert subregion_values[0, 0, 0, 0] == 1
    assert subregion_values[1, 1, 1, 0] == 2
    assert len(subregions_dict) == 3

    mc.scripts.util.mumax3_regions(system)
    subregions = df.Field.from_file("mumax3_regions.omf")
    assert np.allclose(np.unique(subregions.array), [0, 1])
    assert hasattr(system, "region_relator")
    assert system.region_relator == {"": [], "r1": [0], "r2": [1]}

    subregions = {
        "r1": df.Region(p1=(0, 0, 0), p2=(2, 2, 2)),
        "r2": df.Region(p1=(0, 0, 0), p2=(2, 2, 1)),
    }
    mesh = df.Mesh(p1=(0, 0, 0), p2=(2, 2, 2), cell=(1, 1, 1), subregions=subregions)
    system = mm.System(name="test")
    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=1)

    subregion_values, subregions_dict = mc.scripts.util._identify_subregions(system)
    assert subregion_values[0, 0, 0, 0] == 1
    assert subregion_values[1, 1, 1, 0] == 1
    assert len(subregions_dict) == 3

    mc.scripts.util.mumax3_regions(system)
    subregions = df.Field.from_file("mumax3_regions.omf")
    assert np.allclose(np.unique(subregions.array), [0])
    assert hasattr(system, "region_relator")
    assert system.region_relator == {"": [], "r1": [0], "r2": []}


def test_identify_subregions__three_subregions():
    subregions = {
        "r1": df.Region(p1=(0, 0, 0), p2=(2, 2, 1)),
        "r2": df.Region(p1=(0, 0, 1), p2=(2, 2, 2)),
        "r3": df.Region(p1=(0, 0, 2), p2=(2, 2, 3)),
    }
    mesh = df.Mesh(p1=(0, 0, 0), p2=(2, 2, 3), cell=(1, 1, 1), subregions=subregions)
    system = mm.System(name="test")
    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=1)

    subregion_values, subregions_dict = mc.scripts.util._identify_subregions(system)
    assert subregion_values[0, 0, 0, 0] == 1
    assert subregion_values[1, 1, 1, 0] == 2
    assert subregion_values[1, 1, 2, 0] == 3
    assert len(subregions_dict) == 4

    mc.scripts.util.mumax3_regions(system)
    subregions = df.Field.from_file("mumax3_regions.omf")
    assert np.allclose(np.unique(subregions.array), [0, 1, 2])
    assert hasattr(system, "region_relator")
    assert system.region_relator == {"": [], "r1": [0], "r2": [1], "r3": [2]}


def test_mumax3_regions__too_many_ms():
    subregions = {
        "r1": df.Region(p1=(0, 0, 0), p2=(200, 2, 1)),
        "r2": df.Region(p1=(0, 0, 2), p2=(200, 2, 3)),
    }
    mesh = df.Mesh(p1=(0, 0, 0), p2=(200, 2, 3), cell=(1, 1, 1), subregions=subregions)
    system = mm.System(name="test")

    def ms_fun(pos):
        x, y, z = pos
        return x

    system.m = df.Field(mesh, nvdim=3, value=(0, 0, 1), norm=ms_fun)
    with pytest.raises(ValueError):
        mc.scripts.mumax3_regions(system)



import discretisedfield as df
import micromagneticmodel as mm
import numpy as np
import ubermagutil.typesystem as ts

import mumax3c as mc


def energy_script(system, ovf_format, abspath):
    mx3 = ""
    for term in system.energy:
        if isinstance(term, mm.Zeeman):  # Handled separately
            continue
        elif len(system.energy.get(type=type(term))) > 1:
            raise RuntimeError(
                "Mumax3 does not allow more than one energy term of the same class "
                "except for the Zeeman term."
            )
        else:
            mx3 += globals()[f"{term.__class__.__name__.lower()}_script"](
                term, system, ovf_format, abspath
            )

    if zeeman_terms := system.energy.get(type=mm.Zeeman):
        for term in zeeman_terms:
            if isinstance(term.H, (tuple, list, dict, np.ndarray)):
                H_field = df.Field(mesh=system.m.mesh, nvdim=3, value=term.H)
                mx3 += zeeman_script(mm.Zeeman(H=H_field), system, ovf_format, abspath)
            else:
                mx3 += zeeman_script(term, system, ovf_format, abspath)

        mx3 += "tableadd(E_Zeeman)\n"

    # Demagnetisation in mumax3 is enabled by default.
    if mm.Demag() not in system.energy:
        mx3 += "enabledemag = false\n\n"

    return mx3


def exchange_script(term, system, ovf_format, abspath):
    mx3 = "// Exchange energy\n"
    mx3 += mc.scripts.set_parameter(
        parameter=term.A,
        name="Aex",
        system=system,
        ovf_format=ovf_format,
        abspath=abspath,
    )
    mx3 += "tableadd(E_exch)\n"
    return mx3


def zeeman_script(term, system, ovf_format, abspath):
    # mx3 file takes B, not H.
    H = term.H
    if isinstance(H, dict):
        B = dict()
        for key, value in H.items():
            B[key] = np.multiply(value, mm.consts.mu0)
    else:
        B = np.multiply(H, mm.consts.mu0)

    mx3 = "// Zeeman\n"
    mx3 += mc.scripts.set_parameter(
        parameter=B, name="B_ext", system=system, ovf_format=ovf_format, abspath=abspath
    )
    return mx3


def uniaxialanisotropy_script(term, system, ovf_format, abspath):
    mx3 = "// UniaxialAnisotropy\n"
    if not isinstance(term.K, ts.descriptors.Parameter):
        mx3 += mc.scripts.set_parameter(
            parameter=term.K,
            name="Ku1",
            system=system,
            ovf_format=ovf_format,
            abspath=abspath,
        )
    else:
        mx3 += mc.scripts.set_parameter(
            parameter=term.K1,
            name="Ku1",
            system=system,
            ovf_format=ovf_format,
            abspath=abspath,
        )
        mx3 += mc.scripts.set_parameter(
            parameter=term.K2,
            name="Ku2",
            system=system,
            ovf_format=ovf_format,
            abspath=abspath,
        )

    mx3 += mc.scripts.set_parameter(
        parameter=term.u,
        name="anisU",
        system=system,
        ovf_format=ovf_format,
        abspath=abspath,
    )
    mx3 += "tableadd(E_anis)\n"
    return mx3


def demag_script(term, system, ovf_format, abspath):
    # all energy terms must have the same signature
    mx3 = "// Demag\n"
    mx3 += "enabledemag = true\n"
    mx3 += "tableadd(E_demag)\n"
    return mx3


def dmi_script(term, system, ovf_format, abspath):
    if not system.energy.get(type=mm.Exchange):
        raise RuntimeError(
            "In mumax3 DMI cannot be used without exchange. "
            "Solution: define exchange with a negligible A value."
        )
    elif term.crystalclass.lower() in ["t", "o"]:
        param_name = "Dbulk"
        param_val = term.D
    elif term.crystalclass.lower() in ["cnv_z", "cnv"]:
        param_name = "Dind"
        if isinstance(term.D, dict):
            param_val = {sub_reg: -val for sub_reg, val in term.D.items()}
        else:
            param_val = -term.D
        # In mumax3 D = -D for interfacial DMI
    else:
        raise ValueError(
            f"The {term.crystalclass} crystal class is not supported in mumax3."
        )

    mx3 = "// DMI\n"
    mx3 += mc.scripts.set_parameter(
        parameter=param_val,
        name=param_name,
        system=system,
        ovf_format=ovf_format,
        abspath=abspath,
    )
    # In mumax DMI energy is combined with exchange energy
    return mx3


def cubicanisotropy_script(term, system, ovf_format, abspath):
    mx3 = "// CubicAnisotropy\n"
    mx3 += mc.scripts.set_parameter(
        parameter=term.K,
        name="Kc1",
        system=system,
        ovf_format=ovf_format,
        abspath=abspath,
    )
    mx3 += mc.scripts.set_parameter(
        parameter=term.u1,
        name="anisC1",
        system=system,
        ovf_format=ovf_format,
        abspath=abspath,
    )
    mx3 += mc.scripts.set_parameter(
        parameter=term.u2,
        name="anisC2",
        system=system,
        ovf_format=ovf_format,
        abspath=abspath,
    )
    mx3 += "tableadd(E_anis)\n"

    return mx3


# def magnetoelastic_script(term):
#     B1mx3, B1name = mc.scripts.setup_scalar_parameter(term.B1, "mel_B1")
#     B2mx3, B2name = mc.scripts.setup_scalar_parameter(term.B2, "mel_B2")
#     ediagmx3, ediagname = mc.scripts.setup_vector_parameter(term.e_diag, "mel_ediag")
#     eoffdiagmx3, eoffdiagname = mc.scripts.setup_vector_parameter(
#         term.e_offdiag, "mel_eoffdiag"
#     )

#     mx3 = ""
#     mx3 += B1mx3
#     mx3 += B2mx3
#     mx3 += ediagmx3
#     mx3 += eoffdiagmx3
#     mx3 += "# MagnetoElastic\n"
#     mx3 += "Specify YY_FixedMEL {\n"
#     mx3 += f"  B1 {B1name}\n"
#     mx3 += f"  B2 {B2name}\n"
#     mx3 += f"  e_diag_field {ediagname}\n"
#     mx3 += f"  e_offdiag_field {eoffdiagname}\n"
#     mx3 += "}\n\n"

#     return mx3



import pytest

import mumax3c as mc

not_supported_by_mumax = [
    "TestDemag.test_demag_asymptotic_radius",
    "TestEnergy.test_zeeman_zeeman",
    "TestExchange.test_field",
    "TestHysteresisDriver.test_noevolver_nodriver",
    "TestMinDriver.test_evolver_nodriver",
    "TestMinDriver.test_evolver_driver",
    "TestMinDriver.test_wrong_evolver",
    "test_outputstep",
    "TestPrecession.test_field",
    "TestRKKY.test_scalar",
    "TestThreads.test_threads",
    "TestTimeDriver.test_rungekutta_evolver_nodriver",
    "TestTimeDriver.test_euler_evolver_nodriver",
    "TestTimeDriver.test_theta_evolver_nodriver",
    "TestTimeDriver.test_therm_heun_evolver_nodriver",
    "TestTimeDriver.test_noevolver_nodriver_finite_temperature",
    "TestTimeDriver.test_wrong_evolver",
    "TestTimeDriver.test_noevolver_driver",
    "TestZeeman.test_time_vector",
    "TestZeeman.test_time_dict",
    "TestZeeman.test_time_field",
    "TestMinDriver.test_noevolver_driver",
    "TestDMI.test_crystalclass",
    "TestCubicAnisotropy.test_field_field_field",
    "TestDamping.test_field",
    "TestDynamics.test_field_field",
    "TestSlonczewski.test_field_values",
    "TestZhangLi.test_field_scalar",
    "TestUniaxialAnisotropy.test_field_vector",
    "TestUniaxialAnisotropy.test_scalar_field",
    "TestUniaxialAnisotropy.test_field_field",
    "TestUniaxialAnisotropy.test_field_dict",
]

missing_in_mumax3c = [
    "TestCompute.test_energy",
    "TestCompute.test_energy_density",
    "TestCompute.test_effective_field",
    "TestCompute.test_invalid_func",
    "TestCompute.test_dmi",
    "TestCompute.test_slonczewski",
    "TestCompute.test_zhang_li",
    "TestCubicAnisotropy.test_field_vector_vector",
    "TestDamping.test_dict",
    "TestDynamics.test_scalar_dict",
    "TestFixedSubregions.test_fixed_subregions",
    "TestSlonczewski.test_single_values",
    "TestSlonczewski.test_single_values_finite_temperature",
    "TestSlonczewski.test_dict_values",
    "TestZhangLi.test_dict_scalar",
    "TestZhangLi.test_time_scalar_scalar",
    "test_multiple_drives_compute",
    "TestZhangLi.test_scalar_scalar",
    "TestPrecession.test_scalar",
    "TestDMI.test_scalar",
    "TestDMI.test_dict",
    "TestDamping.test_scalar",
    "TestPrecession.test_dict",
]


@pytest.fixture(scope="module")
def calculator():
    return mc


@pytest.fixture(autouse=True)
def skip_unsupported_or_missing(request):
    requesting_test_function = (
        f"{request.cls.__name__}.{request.function.__name__}"
        if request.cls
        else request.function.__name__
    )
    if requesting_test_function in not_supported_by_mumax:
        pytest.skip("Not supported by mumax3.")
    elif requesting_test_function in missing_in_mumax3c:
        pytest.xfail("Currently not implemented in mumax3c.")



from .driver import Driver


class MinDriver(Driver):
    """Energy minimisation driver.

    Only attributes in ``_allowed_attributes`` can be defined. For details on
    possible values for individual attributes and their default values, please
    refer to ``Mumax3`` documentation (https://mumax.github.io).

    Examples
    --------
    1. Defining driver with a keyword argument.

    >>> import mumax3c as mc
    ...
    >>> md = mc.MinDriver(DemagAccuracy=6)

    2. Passing an argument which is not allowed.

    >>> import mumax3c as mc
    ...
    >>> md = mc.MinDriver(myarg=1)
    Traceback (most recent call last):
       ...
    AttributeError: ...

    3. Getting the list of allowed attributes.

    >>> import mumax3c as mc
    ...
    >>> md = mc.MinDriver()
    >>> md._allowed_attributes
    [...]

    """

    _allowed_attributes = [
        "MinimizerStop",
        "DemagAccuracy",
        "Headroom",
        "LastErr",
        "MaxErr",
        "MinimizerSamples",
        "NEval",
        "PeakErr",
    ]

    def _checkargs(self, **kwargs):
        pass  # no kwargs should be checked

    @property
    def _x(self):
        return "t"  # TODO correct iteration



"""Mumax3 calculator."""
import importlib.metadata

import pytest

import mumax3c.mumax3
import mumax3c.scripts

# from .compute import compute  # compute is not yet supported
from .delete import delete
from .drivers import MinDriver, RelaxDriver, TimeDriver

runner = mumax3c.mumax3.Runner()
"""Controls the default runner.

``runner`` gives access to the default mumax3 runner used by ``mumxa3c``. For
details refer to ``mumax3c.mumax3.Runner``.

Examples
--------
``runner.runner``
    Returns the default runner; selects the best available runner if unset. A
    different ``Mumax3Runner`` can be passed to be used instead. The new runner
    is tested first.

``runner.autoselect_runner()``
    Lets ``mumax3c`` select the best runner. Can be used to reset the runner
    after overwriting it manually.

See Also
--------
:py:class:`~mumax3c.mumax3.Runner`

"""


def test():
    """Run all package tests.

    Examples
    --------
    1. Run all tests.

    >>> import mumax3c as mc
    ...
    >>> # mc.test()

    """
    return pytest.main(
        ["-m", "not travis and not docker", "-v", "--pyargs", "mumax3c"]
    )  # pragma: no cover


def test_docker():
    return pytest.main(
        ["-m", "docker", "-v", "--pyargs", "mumax3c"]
    )  # pragma: no cover


__version__ = importlib.metadata.version(__package__)



import contextlib
import itertools
import numbers
import pathlib

import discretisedfield as df
import numpy as np


def _identify_subregions(system):
    subregion_indices = np.zeros((*system.m.mesh.n, 1), dtype=int)
    subregion_dict = {0: ""}
    if system.m.mesh.subregions:
        subregion_dict.update(zip(itertools.count(start=1), system.m.mesh.subregions))
        # Reversed to get same functionality as oommf if subregions overlap
        for sr_index, sr_name in reversed(subregion_dict.items()):
            with contextlib.suppress(KeyError):
                slices = system.m.mesh.region2slices(system.m[sr_name].mesh.region)
                subregion_indices[slices] = sr_index
    return subregion_indices, subregion_dict


def mumax3_regions(system, ovf_format="bin4", abspath=True):
    """Convert ubermag subregions and changing Ms values into mumax3 regions.

    In this method, 'region' refers to mumax3, 'subregion refers to ubermag.

    If ``abspath=True`` use an absolute path for the regions omf file otherwise just the
    filename.

    """
    mx3 = ""
    sr_indices, sr_dict = _identify_subregions(system)

    Ms_array = system.m.norm.array
    if np.any(np.isnan(Ms_array)):  # Not sure about this.
        raise ValueError("Ms values cannot be nan.")
    if 0 in Ms_array:
        region_indices = np.full((*system.m.mesh.n, 1), fill_value=255)
        mx3 += "Msat.setRegion(255, 0.0)\n"
        max_index = 254
    else:
        region_indices = np.empty((*system.m.mesh.n, 1))
        max_index = 255

    # dict.fromkeys(..., []) would use the same list for all items
    region_relator = dict.fromkeys(sr_dict.values())
    for key in region_relator:
        region_relator[key] = []
    unique_index = -1

    for sr_index, sr_name in sr_dict.items():
        for ms in unique_with_accuracy(Ms_array[sr_indices == sr_index]):
            if ms == 0:
                continue
            unique_index += 1
            mx3 += f"Msat.setregion({unique_index}, {ms})\n"
            region_indices[
                (sr_indices == sr_index) & np.isclose(Ms_array, ms)
            ] = unique_index
            region_relator[sr_name].append(unique_index)

    if unique_index > max_index:
        raise ValueError(
            "mumax3 does not allow more than 256 seperate regions to be set. The"
            " number of mumax3 regions is determined by the number of unique"
            " combinations of `discretisedfield` subregions and saturation"
            f" magnetisation values. Found {len(system.m.mesh.subregions)} subregions"
            f" and {len(unique_with_accuracy(Ms_array))} Ms values resulting in"
            f" {unique_index} mumax3 regions."
        )

    region_path = pathlib.Path("mumax3_regions.omf")
    df.Field(system.m.mesh, nvdim=1, value=region_indices).to_file(
        str(region_path), representation=ovf_format
    )
    system.region_relator = region_relator
    if abspath:
        region_path = region_path.absolute().as_posix()  # / as path separator required
    mx3 += f'\nregions.LoadFile("{region_path}")\n\n'
    return mx3


def unique_with_accuracy(array, accuracy=14):
    """Find unique float values with accuracy post-decimal digits.

    The method divides the input by its maximum value to ensure that the values have the
    form 0.xxx. Rounding is then done with ``accuracy`` post-decimal digits.

    """
    if len(array.flat) <= 1:
        return np.array(array.flat)
    array_max = np.max(array)
    return np.unique(np.round(array / array_max, decimals=accuracy)) * array_max


def set_parameter(parameter, name, system, ovf_format="bin4", abspath=True):
    mx3 = ""
    # Spatially constant scalar parameter.
    if isinstance(parameter, numbers.Real):
        mx3 += f"{name} = {parameter}\n"

    # Spatially constant vector parameter.
    elif isinstance(parameter, (list, tuple, np.ndarray)):
        mx3 += "{} = vector({}, {}, {})\n".format(name, *parameter)

    # Spatially varying parameter defined using subregions.
    elif isinstance(parameter, dict):
        if "default" in parameter:
            value = parameter.pop("default")
            if isinstance(value, numbers.Real):
                mx3 += f"{name} = {value}\n"
            elif isinstance(value, (list, tuple, np.ndarray)):
                mx3 += f"{name} = , vector({{}}, {{}}, {{}}))\n".format(*value)
        for key, value in parameter.items():
            if ":" in key:
                mx3 += _set_inter_reg_params(key, value, name, system)
            else:
                for region in system.region_relator[key]:
                    if isinstance(value, numbers.Real):
                        mx3 += f"{name}.setregion({region}, {value})\n"
                    elif isinstance(value, (list, tuple, np.ndarray)):
                        mx3 += (
                            f"{name}.setregion({region}, "
                            "vector({}, {}, {}))\n".format(*value)
                        )

    elif isinstance(parameter, df.Field) and name == "B_ext":
        if file_list := list(pathlib.Path(".").glob("B_ext*.ovf")):
            num_ovf = len(file_list)
            b_ext_path = pathlib.Path(f"B_ext_{num_ovf}.ovf")
            parameter.to_file(b_ext_path, representation=ovf_format)
            if abspath:
                b_ext_path = b_ext_path.absolute().as_posix()  # / as separator required
            mx3 += f'B_ext.add(LoadFile("{b_ext_path}"), 1)\n'
        else:
            b_ext_path = pathlib.Path("B_ext.ovf")
            parameter.to_file("B_ext.ovf", representation=ovf_format)
            if abspath:
                b_ext_path = b_ext_path.absolute().as_posix()  # / as separator required
            mx3 += f'B_ext.add(LoadFile("{b_ext_path}"), 1)\n'  # 1: constant in time

    else:
        # In mumax3, the parameter cannot be set using Field except for Zeeman field.
        msg = f"Cannot use {type(parameter)} to set parameter."
        raise TypeError(msg)

    return mx3


def _set_inter_reg_params(key, value, name, system):
    sub_regions = key.split(":")
    if name not in {"Aex", "Dind"}:  # ext_InterDbulk not available.
        raise ValueError(
            "Only Aex and Dind can be set for different region in mumax3."
            f" Cannot set inter region {name}"
        )
    else:
        for reg_pair in itertools.product(
            system.region_relator[sub_regions[0]], system.region_relator[sub_regions[1]]
        ):
            if name == "Aex":
                return f"\next_InterExchange({reg_pair[0]}, {reg_pair[1]}, {value})\n"
            else:
                return f"\next_InterDind({reg_pair[0]}, {reg_pair[1]}, {value})\n"



import abc
import logging
import pathlib
import shutil
import subprocess as sp
import tempfile
import time

import micromagneticmodel as mm
import ubermagutil as uu

import mumax3c as mc

log = logging.getLogger("mumax3c")


class Mumax3Runner(mm.ExternalRunner):
    """Abstract class for running mumax3."""

    @property
    def package_name(self):
        """Simulation package name."""
        return "mumax3"

    @abc.abstractmethod
    def _call(self, argstr, need_stderr=False, dry_run=False):
        """This method should be implemented in subclass."""

    @property
    def status(self):
        """Run a macrospin example for 1 ps through mumax3c and print the mumax3
        status.

        Returns
        -------
        int

            If ``0``, the mumax3 is found and running. Otherwise, ``1`` is
            returned.

        Examples
        --------
        1. Checking the mumax3 status.

        >>> import mumax3c as mc
        ...
        >>> mc.runner.runner.status
        Running mumax3...
        mumax3 found and running.
        0

        """
        system = mm.examples.macrospin()
        try:
            td = mc.TimeDriver()
            td.drive(system, t=1e-12, n=1, runner=self)
            print("mumax3 found and running.")
            return 0
        except (EnvironmentError, RuntimeError):
            print("Cannot find mumax3.")
            return 1


@uu.inherit_docs
class ExeMumax3Runner(Mumax3Runner):
    """mumax3 runner using mumax3 executable, which can be found on $PATH.

    Parameters
    ----------
    mumax3_exe: str

        Name or path of the mumax3 executable. Defaults to
        ``mumax3``.

    """

    def __init__(self, mumax3_exe="mumax3"):
        if isinstance(mumax3_exe, str):
            mumax3_exe = [mumax3_exe]
        self.mumax3_exe = mumax3_exe

    def _call(self, argstr, need_stderr=False, dry_run=False):
        cmd = self.mumax3_exe + [argstr]
        if dry_run:
            return " ".join(cmd)
        else:
            return sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)


class Runner:
    """Control the default runner.

    Parameters
    ----------
    cache_runner : bool
        If ``True`` the best way to run mumax3 is only determined once and the
        result is cached. Subsequent calls to the property ``runner`` will
        return the ``Mumax3Runner`` object from the cache. Setting this
        parameter to ``False`` will cause it to check for available methods
        again. Defaults to ``True``.

    mumax3_exe : str

        The name or path of the executable ``mumax3`` command. Defaults to
        ``mumax3``.

    optirun_exe : str

        The name or path of the executable ``optirun`` command. Defaults to
        ``optirun``.

    """

    def __init__(self):
        self.cache_runner = True
        self.mumax3_exe = "mumax3"
        self.optirun_exe = "optirun"
        self._runner = None

    @property
    def runner(self):
        """Return the default mumax3 runner.

        The default runner is determined using ``autoselect_runner()``. If
        ``cache_runner`` is ``True`` the runner is cached during the first call
        and the same runner is returned in subsequent calls to this property.

        This property also allows to set a specific ``Mumax3Runner``. Before
        setting, a new runner is first checked to be functional by calling
        ``runner.status``.

        Examples
        --------
        1. Getting mumax3 Runner.

        >>> import mumax3c as mc
        ...
        >>> runner = mc.runner.runner
        >>> isinstance(runner, mc.mumax3.Mumax3Runner)
        True
        """
        if self.cache_runner and self._runner is not None:
            log.debug("Returning cached runner.")
            return self._runner
        self.autoselect_runner()
        return self._runner

    @runner.setter
    def runner(self, runner):
        if runner.status != 0:
            raise ValueError(f"{runner=} cannot be used.")
        self._runner = runner

    def autoselect_runner(self):
        """Find the best available way to run mumax3.

        The method tries to find a suitable runner by checking the availability of
        ``optirun`` and ``mumax3`` in this order. If no runner can be found an
        ``EnvironmentError`` is raised.

        Raises
        ------
        EnvironmentError

            If mumax3 cannot be found on host.

        Examples
        --------
        1. Getting mumax3 Runner.

        >>> import mumax3c as mc
        ...
        >>> mc.runner.autoselect_runner()
        >>> runner = mc.runner.runner
        >>> isinstance(runner, mc.mumax3.Mumax3Runner)
        True

        """
        log.debug(
            "Starting autoselect_runner: cache_runner=%(cache_runner)s, "
            "mumax3_exe=%(mumax3_exe)s, optirun_exe=%(optirun_exe)s, ",
            {
                "cache_runner": self.cache_runner,
                "mumax3_exe": self.mumax3_exe,
                "optirun_exe": self.optirun_exe,
            },
        )

        cmd = []
        log.debug("Step 1: Checking for optirun.")
        optirun_exe = shutil.which(self.optirun_exe)
        log.debug(
            "Output from 'which optirun_exe=%(optirun_exe)s",
            {"optirun_exe": optirun_exe},
        )
        if optirun_exe:
            cmd.append(self.optirun_exe)

        log.debug("Step 2: Checking for mumax3")
        mumax3_exe = shutil.which(self.mumax3_exe)
        log.debug(
            "Output from 'which mumax3_exe=%(mumax3_exe)s", {"mumax3_exe": mumax3_exe}
        )
        if mumax3_exe:
            cmd.append(self.mumax3_exe)
            self._runner = ExeMumax3Runner(cmd)
        else:
            msg = (
                "Mumax3 cannot be found. Mumax3 does not come automatically with"
                " Ubermag. Make sure it is properly installed and can be found on the"
                " command line."
            )
            raise EnvironmentError(msg)


def overhead():
    """Run a macrospin example for 1 ps through ``mumax3c`` and directly and
    return the difference in run times.

    Returns
    -------
    float

      The time difference (overhead) between running mumax3 though ``mumax3c``
      and directly.

    Examples
    --------
    1. Getting the overhead time.

    >>> import mumax3c as mc
    ...
    >>> isinstance(mc.mumax3.overhead(), float)
    Running mumax3...
    True

    """
    with tempfile.TemporaryDirectory() as workingdir:
        with uu.changedir(workingdir):
            # Running mumax3 through mumax3c.
            system = mm.examples.macrospin()
            td = mc.TimeDriver()
            mumax3c_start = time.time()
            td.drive(system, t=1e-12, n=1)
            mumax3c_stop = time.time()
            mumax3c_time = mumax3c_stop - mumax3c_start

            # Running mumax3 directly.
            mumax3_runner = mc.runner.runner
            mx3path = pathlib.Path(f"{system.name}/drive-0/macrospin.mx3").resolve()
            mumax3_start = time.time()
            mumax3_runner.call(str(mx3path))
            mumax3_stop = time.time()
            mumax3_time = mumax3_stop - mumax3_start

    return mumax3c_time - mumax3_time



import abc
import pathlib

import discretisedfield as df
import micromagneticmodel as mm
import ubermagtable as ut
import ubermagutil as uu

import mumax3c as mc


class Driver(mm.ExternalDriver):
    """Driver base class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if hasattr(self, "evolver"):
            self.autoselect_evolver = False
        else:
            self.autoselect_evolver = True

    @abc.abstractmethod
    def _checkargs(self, **kwargs):
        """Abstract method for checking arguments."""

    def drive_kwargs_setup(self, drive_kwargs):
        """Additional keyword arguments allowed for drive.

        This function tests additional keyword arguments that have been passed to the
        ``drive`` method (it is not intended for direct use). A drive in mumax3c can
        accept the following additional keyword arguments.

        Parameters
        ----------
        abspath : bool, optional

            If ``True`` absolute paths for all input files (e.g. the initial
            magnetisation) are used in the mx3 file. If ``False`` relative paths are
            used. Defaults to ``True``.

        """
        self._checkargs(**drive_kwargs)
        drive_kwargs.setdefault("abspath", True)

        # TODO OOMMF support additional arguments; are there equivalent options in mumax
        # fixed_subregions = None  # mumax?
        # compute = None  # mumax?
        # output_step = False  # mumax?

    def schedule_kwargs_setup(self, schedule_kwargs):
        """Additional keyword arguments allowed for schedule.

        This function tests additional keyword arguments that have been passed to the
        ``schedule`` method (it is not intended for direct use). A drive in oommfc can
        accept the following additional keyword arguments.

        It is the user's responsibility to ensure that OOMMF can be executed from the
        scheduled job.

        Parameters
        ----------
        abspath : bool, optional

            If ``True`` absolute paths for all input files (e.g. the initial
            magnetisation) are used in the mx3 file. If ``False`` relative paths are
            used. Defaults to ``True``.

        """
        self._checkargs(**schedule_kwargs)
        schedule_kwargs.setdefault("abspath", True)

    def _write_input_files(self, system, **kwargs):
        self.write_mx3(system, **kwargs)

    def write_mx3(self, system, dirname=".", ovf_format="bin8", abspath=True, **kwargs):
        """Write the mx3 file and related files.

        Takes ``micromagneticmodel.System`` and write the mx3 file (and related files)
        to drive it in the phase space. The files are written directly to directory
        ``dirname`` (if not specified the current working directory). No additional
        subdirectiories are created.

        This method accepts any other arguments that could be required by the
        specific driver.

        Users are generally not encouraged to use this method directly. Instead
        ``Driver.drive``, ``Driver.schedule``, or ``mc.schedule`` should be used to
        write the files an run the simulation. This method is provided to give advanced
        users full flexibility.

        Parameters
        ----------
        system : micromagneticmodel.System

            System object to be driven.

        dirname : str, optional

            Name of a directory in which the input files are stored.
            If not specified the current workinng
            directory is used.

        ovf_format : str

            TODO UPDATE TO THE SUPPORTED MUMAX FILE TYPES
            Format of the magnetisation output files written by mumax3. Can be
            one of ``'bin8'`` (binary, double precision), ``'bin4'`` (binary,
            single precision) or ``'txt'`` (text-based, double precision).
            Defaults to ``'bin8'``.

        abspath : bool, optional

            If ``abspath=True`` absolute paths for additional input files (e.g. initial
            magnetisation) are written to the mx3 file. If ``abspath=False`` only
            filenames are written to the file. Relative files require mumax3 to be run
            from inside the directory containing the mx3 and other input files.

        """
        with uu.changedir(dirname):
            mx3 = mc.scripts.system_script(
                system, ovf_format=ovf_format, abspath=abspath
            )
            mx3 += mc.scripts.driver_script(
                self,
                system,
                compute=None,  # TODO does mumax3 support compute?
                ovf_format=ovf_format,
                **kwargs,
            )
            with open(self._mx3filename(system), "wt", encoding="utf-8") as mx3file:
                mx3file.write(mx3)

            # Generate and save json info file for a drive (not compute).
            if True:  # compute is None:  # TODO does mumxa3 support compute?
                self._write_info_json(system, **kwargs)

        # TODO if self/system is modified for mx3 creation reset it here
        delattr(system, "region_relator")

    def _call(self, system, runner, verbose=1, dry_run=False, **kwargs):
        if runner is None:
            runner = mc.runner.runner
        if dry_run:
            return runner.call(argstr=self._mx3filename(system), dry_run=True)
        else:
            runner.call(
                argstr=self._mx3filename(system),
                verbose=verbose,
                total=kwargs.get("n"),
                glob_name=f"{system.name}.out/m_full*.ovf",
            )

    def _schedule_commands(self, system, runner):
        if runner is None:
            runner = mc.runner.runner
        return [
            runner._call(argstr=self._mx3filename(system), dry_run=True),
        ]

    def _read_data(self, system):
        # Update system's magnetisation. An example .ovf filename: m_full000000.ovf
        ovffiles = pathlib.Path(f"{system.name}.out").glob("m_full*.ovf")
        lastovffile = sorted(ovffiles)[-1]
        # pass Field.array instead of Field for better performance
        # Mumax3 norm changes so need to set back to old norm
        norm_field = system.m.norm
        system.m.array = df.Field.from_file(str(lastovffile)).array
        system.m.norm = norm_field

        system.table = ut.Table.fromfile(
            str(pathlib.Path(f"{system.name}.out/table.txt")), x=self._x
        )

    @staticmethod
    def _mx3filename(system):
        return f"{system.name}.mx3"



from .driver import driver_script
from .energy import energy_script
from .magnetisation import magnetisation_script
from .mesh import mesh_script
from .system import system_script
from .util import mumax3_regions, set_parameter



import os
import shutil

import mumax3c as calc


def check_runner(runner):
    # Testing mumax3c on an mx3 file to make it independent of mumax3c.
    dirname = os.path.join(os.path.dirname(__file__), "test_sample")
    os.chdir(dirname)
    argstr = "test_mumax3.mx3"

    res = runner.call(argstr)
    assert res.returncode == 0

    # Cleanup created files.
    for f in os.listdir(dirname):
        if f.endswith(".json"):
            os.remove(os.path.join(dirname, f))
        elif os.path.isdir(f) and f.endswith(".out"):
            shutil.rmtree(f)


def test_exemumax3runner():
    # ExeMumax3Runner runs when callable mumax3 exists ('mumax3').
    mumax3_exe = "mumax3"
    runner = calc.mumax3.ExeMumax3Runner(mumax3_exe)
    check_runner(runner)


def test_get_mumax3_runner():
    # Tclmumax3Runner
    runner = calc.runner.runner
    assert isinstance(runner, calc.mumax3.ExeMumax3Runner)
    check_runner(runner)


def test_status():
    assert calc.runner.runner.status == 0


def test_overhead():
    assert isinstance(calc.mumax3.overhead(), float)



import mumax3c as mc


def system_script(system, ovf_format, abspath=True, **kwargs):
    if ovf_format in ["bin4", "bin8"]:
        ovf_format = "bin4"  # mumax3 uses single precision
        output_format = "OVF2_BINARY"
    elif ovf_format == "txt":
        output_format = "OVF2_TEXT"
    else:
        raise ValueError(f"Invalid {ovf_format=}.")

    mx3 = ""
    # Output options
    mx3 += f"OutputFormat = {output_format}\n\n"
    # Mesh and energy scripts.
    mx3 += mc.scripts.mesh_script(system)
    mx3 += mc.scripts.magnetisation_script(
        system, ovf_format=ovf_format, abspath=abspath
    )
    mx3 += mc.scripts.energy_script(system, ovf_format=ovf_format, abspath=abspath)

    return mx3
